from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from loguru import logger
from pydantic import BaseModel


class Feedback(BaseModel):
    """Dataclass for feedback given by a user from a UI component."""

    feedback_type: Optional[str]
    metadata: Dict[str, Union[List[Any], float, int, str, dict]]

    def save_local(self, path: str, file_name: Optional[str] = None):
        if path is None:
            raise TypeError("Specify the local path where you would like to save your Trubric json.")
        if file_name is None:
            file_name = f"{self.feedback_type}.json"
        with open(Path(path) / file_name, "w") as file:
            file.write(self.json(indent=4))
            logger.info(f"Trubric saved to {Path(path) / file_name}.")



import streamlit as st

from trubrics.context import DataContext
from trubrics.example import get_titanic_data_and_model
from trubrics.example import titanic_config as tc
from trubrics.feedback import FeedbackCollector

st.set_page_config(page_title="User Feedback Titanic", page_icon="assets/logo-favicon.png")


# @st.cache
def init_data():
    train_df, test_df, model = get_titanic_data_and_model()

    # init data context
    data_context = DataContext(
        training_data=train_df,
        testing_data=test_df,
        target="Survived",
        categorical_columns=tc.CATEGORICAL_COLUMNS,
        business_columns=tc.BUSINESS_COLUMNS,
    )
    return data_context, model


data_context, model = init_data()
collector = FeedbackCollector(data=data_context, model=model)

with st.sidebar:
    st.title("Modify features to test the model...")
    collector.generate_what_if()

st.title("View model prediction")
if collector.what_if_prediction:
    prediction = '<p style="color:Green;">This passenger would have survived.</p>'
else:
    prediction = '<p style="color:Red;">This passenger would have died.</p>'
st.markdown(prediction, unsafe_allow_html=True)

st.title("Send model feedback")
collector.save_feedback(path=".", file_name="feedback.json")

st.title("View data")
data_view = st.selectbox(label="", options=("View full test set", "View test set errors", "View split by target"))
if data_view == "View full test set":
    st.dataframe(collector.trubrics_model.data.renamed_testing_data)
elif data_view == "View test set errors":
    st.dataframe(collector.trubrics_model.testing_data_errors.rename(collector.trubrics_model.data.business_columns))
elif data_view == "View split by target":
    target_split = st.radio(label="", options=("survived", "died"))
    if target_split == "survived":
        st.dataframe(
            collector.trubrics_model.data.testing_data.loc[lambda x: x[collector.trubrics_model.data.target] == 0]
        )
    elif target_split == "died":
        st.dataframe(
            collector.trubrics_model.data.testing_data.loc[lambda x: x[collector.trubrics_model.data.target] == 1]
        )
    else:
        raise NotImplementedError()

else:
    raise NotImplementedError("The data_view must be one of the selectbox options")



from sklearn.metrics import accuracy_score, make_scorer


def custom_binary_error(y_true, y_pred):
    """
    An example of an error metric for binary classification.
    """
    accuracy = accuracy_score(y_true, y_pred)
    return round(1 - accuracy, 3)


custom_score = make_scorer(custom_binary_error, greater_is_better=False)
custom_scorers = {"my_custom_loss": custom_score}



from typing import Any, Dict, Optional

from pydantic import BaseModel, validator
from sklearn.metrics._scorer import _BaseScorer

from trubrics.context import DataContext
from trubrics.exceptions import UnknownValidationError
from trubrics.validations import ModelValidator, Trubric


class TrubricRun(BaseModel):
    """The TrubricRun object to group all necessary code for a run. Load data and models from
    remote locations or locally for validation within a pipeline.

    Attributes:
        data_context: a data context to validate a model on
        model: a model to validate
        trubric: a Trubric object listing all validations to execute
        custom_validator: an optional custom validator
        custom_scorers: an optional dict of custom scorers for computing custom metrics
        slicing_functions: an optional dict of slicing functions
    """

    data_context: DataContext
    model: Any
    trubric: Trubric
    custom_validator: Optional[Any] = None
    custom_scorers: Optional[Dict[str, Any]] = None
    slicing_functions: Optional[Dict[str, Any]] = None

    @validator("custom_validator")
    def custom_validator_inherits_validator(cls, val):
        if issubclass(val, ModelValidator):
            return val
        raise TypeError("Wrong type for 'custom_validator', must be subclass of ModelValidator.")

    @validator("custom_scorers")
    def custom_scorer_is_make_scorer(cls, val):
        for scorer in val:
            if not issubclass(type(val[scorer]), _BaseScorer):
                raise TypeError("Each scorer must be subclass of scikit-learn's _BaseScorer.")
        return val


def run_trubric(tr: TrubricRun):
    if tr.custom_validator is not None:
        model_validator = tr.custom_validator(
            data=tr.data_context,
            model=tr.model,
            custom_scorers=tr.custom_scorers,
            slicing_functions=tr.slicing_functions,
        )
    else:
        model_validator = ModelValidator(
            data=tr.data_context,
            model=tr.model,
            custom_scorers=tr.custom_scorers,
            slicing_functions=tr.slicing_functions,
        )
    for validation in tr.trubric.validations:
        args = validation.validation_kwargs["args"]
        kwargs = validation.validation_kwargs["kwargs"]
        try:
            validation_result = getattr(model_validator, validation.validation_type)(*args, **kwargs)
            new_validation = validation.copy()
            new_validation.outcome = validation_result.outcome
            new_validation.result = validation_result.result
            yield new_validation
        except AttributeError:
            raise UnknownValidationError(
                f"The validation '{validation.validation_type}' does not appear to belong to a validator."
                " Try adding the object that generated the validation to the 'custom_validator' parameter."
            )



from examples.classification_titanic.custom_scorer import custom_scorers
from examples.classification_titanic.custom_validator import CustomValidator
from examples.classification_titanic.slicing_functions import slicing_functions
from trubrics.context import DataContext
from trubrics.example.titanic import get_titanic_data_and_model
from trubrics.validations import Trubric
from trubrics.validations.run import TrubricRun

train_df, test_df, model = get_titanic_data_and_model()

data_context = DataContext(
    testing_data=test_df,
    minimum_functionality_data=test_df.iloc[:5],
    training_data=train_df,
    target="Survived",
)
trubric = Trubric.parse_file("examples/classification_titanic/my_first_trubric.json")

RUN_CONTEXT = TrubricRun(
    data_context=data_context,
    model=model,
    trubric=trubric,
    custom_validator=CustomValidator,
    custom_scorers=custom_scorers,
    slicing_functions=slicing_functions,
)
